/* This file is part of Cloudy and is copyright (C)1978-2008 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */

#ifndef _LINES_SERVICE_H_
#define _LINES_SERVICE_H_


/**outline - adds line photons to reflin and outlin */
void outline( transition *t );

/** enter lines into the line storage array, called once per zone for each line
\param xInten xInten - local emissivity per unit vol, no fill fac
\param wavelength lam integer wavelength
\param *chLab string label for ion
\param chInfo character type of entry for line - 'c' cooling, 'h' heating, 'i' info only, 'r' recom line
\param *chComment string explaining line 
*/
void linadd(
  double xInten,
  realnum wavelength,
  const char *chLab,
  char chInfo ,	
  const char *chComment );

/** put forbidden line into stack, using index derived below 
\param xInten xInten - local emissivity per unit vol, no fill fac
\param wavelength lam integer wavelength
\param *chLab string label for ion
\param ipnt offset of line in continuum mesh
\param chInfo character type of entry for line - 'c' cooling, 'h' heating, 'i' info only, 'r' recom line
\param lgOutToo should line be included in outward beam?
\param *chComment string explaining line 
*/
void lindst(double xInten, 
  realnum wavelength, 
  const char *chLab, 
  long int ipnt, 
  char chInfo, 
  bool lgOutToo,
  const char *chComment);

/** absorption due to continuous opacity 
\param emissivity [erg cm-3 s-1] in inward direction 
\param emissivity [erg cm-3 s-1] in outward direction 
\param array index for continuum frequency 
*/
double emergent_line( 
	 /* emissivity [erg cm-3 s-1] in inward direction */
	 double emissivity_in , 
	 /* emissivity [erg cm-3 s-1] in outward direction */
	 double emissivity_out , 
	 /* array index for continuum frequency */
	 long int ipCont );

/**PntForLine generate pointer for forbidden line 
\param wavelength wavelength of line in Angstroms 
\param *chLabel label for the line
\param *ipnt this is array index on the f, not c scale,
   for the continuum cell holding the line
*/
void PntForLine(double wavelength, 
  const char *chLabel, 
  long int *ipnt);

/**PutLine enter local line intensity into the intensity stack for eventual printout 
\param *t transition structure for line
\param *chComment a description of the line
*/
void PutLine(transition * t, const char *chComment);

/**PutLine enter local line intensity into the intensity stack for eventual printout 
\param *t transition structure for line
\param *chComment a description of the line
\param *chLabel the line label
*/
void PutLine(transition * t, const char *chComment, const char *chLabel);

/**TexcLine derive excitation temperature of line from contents of line array 
\param *t
*/
double TexcLine(transition * t);

/**DumpLine print various information about an emission line vector, used in debugging 
\param *t
*/
void DumpLine(transition * t);

/**GetGF convert Einstein A into oscillator strength 
\param eina
\param enercm
\param gup
*/
double GetGF(double eina, 
	  double enercm, 
	  double gup);

/**eina convert a gf into an Einstein A 
\param gf
\param enercm
\param gup
*/
double eina(double gf, 
	  double enercm, 
	  double gup);

/**abscf convert gf into absorption coefficient 
\param gf
\param enercm
\param gl
*/
double abscf(double gf, 
	  double enercm, 
	  double gl);

/** setting true will use low-density Lyman branching ratios */
#define LOWDEN_LYMAN 0

/** returns fraction of populations the produce emission 
\param *t
*/
double emit_frac( transition *t );

/** generate null terminated line label from contents of line trans array 
\param *t
*/
void chIonLbl( char*, transition * t );

/**chLineLbl use information in line transfer arrays to generate a line label<BR>
 this label is null terminated 
 \param *t
 */
char* chLineLbl(transition * t );

/**RefIndex calculates the index of refraction of air using the line energy in wavenumbers,
 * used to convert vacuum wavelengths to air wavelengths. 
 \param EnergyWN
 */
double RefIndex(double EnergyWN);


/**PutCS enter a collision strength into an individual line struc 
\param cs
\param *t  the line struc 
*/
void PutCS(double cs, 
  transition * t);

/**WavlenErrorGet - given the real wavelength in A for a line
 * routine will find the error expected between the real 
 * wavelength and the wavelength printed in the output, with 4 sig figs,
  \param wavelength
  \return function returns difference between exact and 4 sig fig wl, so 
  we have found correct line is fabs(d wl) < return 
  */
realnum WavlenErrorGet( realnum wavelength );

/**OccupationNumberLine - derive the photon occupation number at line center for any line 
\param *t
*/
double OccupationNumberLine( transition * t );

/**PutExtra enter and 'extra' intensity source for some line 
\param Extra
*/
void PutExtra(double Extra);

/**TransitionJunk set all elements of transition struc to dangerous values  
\param *t
*/
void TransitionJunk( transition *t );

/**EmLineJunk set all elements of emission struc to dangerous values 
\param *t
*/
void EmLineJunk( emission * t );

/**CollisionJunk set all elements of emission struc to dangerous values  
\param *t
*/
void CollisionJunk( collision * t );

/**StateJunk set all elements of quantumState struc to dangerous values  
\param *t
*/
void StateJunk( quantumState * t );

/**TransitionZero set all elements of transition struc to zero 
\param *t
*/
void TransitionZero( transition *t );

/**EmLineZero set all elements of emission struc to zero 
\param *t
*/
void EmLineZero( emission * t );

/**CollisionZero set all elements of collision struc to zero 
\param *t
*/
void CollisionZero( collision * t );

/**StateZero set all elements of quantumState struc to zero 
\param *t
*/
void StateZero( quantumState * t );

/** convert down coll rate back into electron cs in case other parts of code need this for reference 
\param *t - line struct collision strength is stored in t->cs 
\param rate - deexcitation rate, units s-1 
*/
void LineConvRate2CS( transition * t , realnum rate );

/** convert down coll rate back into electron cs in case other parts of code need this for reference 
\param gHi - stat weight of upper level 
\param rate - deexcitation rate, units s-1 
*/
double ConvRate2CS( realnum gHi , realnum rate );

/**lgTauGood returns true is we have not overrun optical depth scale 
\param *t
*/
bool lgTauGood( transition * t);

/**MakeCS compute collision strength by g-bar approximations 
\param *t
*/
void MakeCS(transition * t );

/**totlin sum total intensity of cooling, recombination, or intensity lines 
\param chInfo chInfor is 1 char,<BR>
	'i' information, <BR>
	'r' recombination or <BR>
	'c' collision
*/
double totlin(
	int chInfo);


/**FndLineHt search through line heat arrays to find the strongest heat source 
\param *level
\param *ipStrong  this is the index of the strongest line in the array on the c scale 
\param *Strong
*/
void FndLineHt(long int *level, 
  /**/
  long int *ipStrong, 
  double *Strong);

/**AddState2Stack add generic quantum state to GenericStates and return pointer to that state. */
quantumState *AddState2Stack( void );

/** \todo 2 bring these two together. */
/**AddLine2Stack add generic emission line to GenericLines and return pointer to that state. */
emission *AddLine2Stack( bool lgRadiativeTrans  );
emission *AddLine2Stack( realnum Aul, transition *trans );


#endif /* _LINES_SERVICE_H_ */
