/* This file is part of Cloudy and is copyright (C)1978-2008 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
/*ParseAtomISO parse information from the atom XX-like command line */
#include "cddefines.h"
#include "elementnames.h"
#include "helike_recom.h"
#include "hydrogenic.h"
#include "iso.h"
#include "parse.h"
#include "phycon.h"
#include "rfield.h"
#include "taulines.h"
#include "thirdparty.h"

/*ParseAtomISO parse parameters off the XX-like command */
void ParseAtomISO(long ipISO, char *chCard )
{
	bool lgEOL;
	long int i, 
		numLevels,
		nelem;

	DEBUG_ENTRY( "ParseAtomISO()" );

	/* look for the name of an element - if we don't find one do the entire
	 * iso sequence - returns negative number if element not found */
	nelem = GetElem( chCard );

	/* H-like Helium is not possible */
	if( ipISO==ipHE_LIKE && nelem==ipHYDROGEN )
	{
		fprintf(ioQQQ," Sorry, H-like He does not exist.\n");
		cdEXIT(EXIT_FAILURE);
	}

	/* collisions - don't pick up the levels collapsed command */
	if( nMatch("COLL",chCard) && !nMatch("LEVE" , chCard ) )
	{
		/* option to turn collisions off, all are
		 * set to 1 in zero.  command can accept only one option at a time */
		if( nMatch("EXCI",chCard) )
		{
			/* turn off collisional excitation */
			iso.lgColl_excite[ipISO] = false;
			phycon.lgPhysOK = false;
		}
		else if( nMatch("IONI",chCard) )
		{
			/* turn off collisional ionization */
			iso.lgColl_ionize[ipISO] = false;
			phycon.lgPhysOK = false;
		}

		else if( nMatch("2S2P",chCard) || ( nMatch("2P2S",chCard) && ipISO == ipH_LIKE ) )
		{
			/* >>chng 02 feb 07, change from 2s2p to l-mixing */
			/* this is the atom h-like collision l-mixing command */
			fprintf(ioQQQ,"This command changed to ATOM H-LIKE COLLISIONS L-MIXING\n");
			fprintf(ioQQQ,"I will parse it for now, but may not in the future.\n");
			/* turn off 2s - 2p collisions */
			iso.lgColl_l_mixing[ipISO] = false;
			phycon.lgPhysOK = false;
		}

		else if( nMatch("L-MI",chCard) )
		{
			if( ipISO == ipH_LIKE )
			{
				/* >>chng 02 feb 07, change from 2s2p to l-mixing */
				/* this is the atom h-like collision l-mixing command */
				iso.lgColl_l_mixing[ipISO] = false;
				phycon.lgPhysOK = false;
			}
			else if( nMatch("THER",chCard) )
			{
				/* use l-mix from 
				 *>>refer	l-mix	all Vrinceanu, D. & Flannery, M. R. 2001, PhysRevA 63, 032701 */
				if( nMatch("NO T",chCard) )			
				{
					/* This is the "NO Thermal average" command.  It
					 * causes collisions strengths to be evaluated at kT rather than
					 * integrated over a Maxwellian peaked at kT. */
					iso.lgCS_therm_ave[ipISO] = false;
				}
				else
				{
					iso.lgCS_therm_ave[ipISO] = true;
				}
			}
			else if( nMatch("PENG",chCard) )
			{
				iso.lgCS_Vrinceanu[ipISO] = false;
			}
			else if( nMatch(" OFF" , chCard ) )
			{
				/* this is the atom xx-like collision l-mixing command */
				/* turn off same-n collisions */
				iso.lgColl_l_mixing[ipISO] = false;
				phycon.lgPhysOK = false;
				iso.lgCS_Vrinceanu[ipISO] = false;
				iso.lgCS_therm_ave[ipISO] = false;
			}
			else
			{
				fprintf( ioQQQ, " needs parameter\n" );
				cdEXIT(EXIT_FAILURE);
			}
		}
		else if( nMatch(" OFF" , chCard ) )
		{
			/* turn everything off, since no keyword given */
			iso.lgColl_excite[ipISO] = false;
			iso.lgColl_ionize[ipISO] = false;
			iso.lgColl_l_mixing[ipISO] = false;
			phycon.lgPhysOK = false;
		}
		else
		{
			fprintf( ioQQQ, " needs parameter\n" );
			cdEXIT(EXIT_FAILURE);
		}
	}

	else if( nMatch("DAMP",chCard) )
	{
		if( ipISO == ipHE_LIKE )
		{
			fprintf(ioQQQ," Sorry, the DAMPING option is not implemented for the he-like sequence.\n");
			cdEXIT(EXIT_FAILURE);
		}

		/* turn off absorption due to Ly alpha damping wings */
		hydro.DampOnFac = 0.;
	}

	else if( nMatch("DIEL",chCard) )
	{
		if( ipISO == ipH_LIKE )
		{
			fprintf(ioQQQ," Sorry, but dielectronic recombination onto the h-like sequence is not possible.\n");
			cdEXIT(EXIT_FAILURE);
		}

		/* This sets which set of data to use for dielectronic recombination.	*/
		if( nMatch(" OFF",chCard) )
		{
			iso.lgDielRecom[ipISO] = false;
		}
		else 
			iso.lgDielRecom[ipISO] = true;
	}

	else if( nMatch("LEVE",chCard) )
	{
		/* the number of levels read in is n, the principal quantum number
		 * only lines with upper levels less than n will be printed */

		/* number of levels for iso-sequence */
		/* there are two options here,
		 * when keyword ELEMENT appears, scan off element name and change levels only for
		 * that one.
		 * when there is no ELEMENT then set all in iso to same number */

		/* lgHydroMalloc is false at start of calculation, set true when space 
		 * allocated for the hydrogen and helium lines.  Once done we must ignore all 
		 * future changes in the number of levels */
		if( nMatch("PRIN",chCard) )
		{
			/* only print - do not change levels */
			iso.lgPrintNumberOfLevels = true;
		}
		else if( !lgHydroMalloc )
		{
			i = 5;
			numLevels = (long int)FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL);

			if( lgEOL )
			{
				/* must be a number, or a key, either large (or limit) or small
				 * if no number but LARGER or SMALL then set to preset number */
				if( nMatch("LARG",chCard) )
				{
					/* there is no real limit, but this includes all levels with tabulated rec coefficient */
					numLevels = RREC_MAXN;
				}

				/* this is small or compact keyword */
				else if( nMatch("SMAL",chCard) || nMatch("COMP",chCard) )
				{
					if( ipISO == ipH_LIKE )
						numLevels = 5;
					else if( ipISO == ipHE_LIKE )
						numLevels = 3;
				}

				else
					/* punch out if no number */
					NoNumb(chCard);
			}
			else if( ( numLevels < 3 ) && !nMatch("COLL",chCard) && ipISO == ipH_LIKE )
			{
				/* only enforce this limit if not working on collapsed levels */
				fprintf( ioQQQ, " cannot have fewer than 3 levels, the requested number was %li\n" ,
					numLevels  );
				fprintf( ioQQQ, " Sorry.\n" );
				cdEXIT(EXIT_FAILURE);
			}
			else if( ( numLevels < 3 ) && !nMatch("COLL",chCard) && ipISO == ipHE_LIKE)
			{
				/* only enforce this limit if not working on collapsed levels */
				fprintf( ioQQQ, " cannot have fewer than 3 levels, the requested number was %li\n" ,
					numLevels  );
				fprintf( ioQQQ, " Sorry.\n" );
				cdEXIT(EXIT_FAILURE);
			}

			if( ipISO == ipH_LIKE && numLevels > NHYDRO_MAX_LEVEL-2 )
			{
				fprintf( ioQQQ, " Not possible to set nhlvl to >NHYDRO_MAX_LEVEL-2= %i\n", 
				  NHYDRO_MAX_LEVEL-2 );
				fprintf( ioQQQ, " change NHYDRO_MAX_LEVEL\n");
				cdEXIT(EXIT_FAILURE);
			}

			/* this is check that alpha transition of highest level is within energy
			 * bounds of continuum */
			if( ipISO == ipH_LIKE &&
				( 2. / POW3((double)numLevels) < rfield.emm ) )
			{
				fprintf( ioQQQ, " Not possible to set iso.numLevels_max[ipH_LIKE][ipHYDROGEN] to such a high value, since "
					"alpha transition not within energy bounds of code\n");

				fprintf( ioQQQ, " lowest energy is %e and corresponding highest level is %li\n" ,
					rfield.emm, (long)pow(2./rfield.emm, 0.3333) );
				cdEXIT(EXIT_FAILURE);
			}

			if( nMatch("COLL",chCard) )
			{
				if( numLevels < 1 )
				{
					fprintf( ioQQQ, "There must be at least one collapsed level.\n");
					cdEXIT(EXIT_FAILURE);
				}

				/* set collapsed levels for entire sequence or just one element */
				if( nMatch("ELEM",chCard) )
				{
					/* check which element is on the line, nelem is atomic number on C scale */
					nelem = GetElem(chCard);
					iso.nCollapsed_max[ipISO][nelem] = numLevels;
					iso_update_num_levels( ipISO, nelem );
				}
				else
				{
					/* keyword element did not appear, so set all iso species to this number */
					for( nelem=ipISO; nelem<LIMELM; ++nelem )
					{
						iso.nCollapsed_max[ipISO][nelem] = numLevels;
						iso_update_num_levels( ipISO, nelem );
					}
				}
			}

			/* we now have the desired number of levels, and it has been fully qualified. 
			 * now check whether ELEMENT keyword is on line, if not then set all to
			 * the number, if so then only element in use */
			else if( nMatch("ELEM",chCard) )
			{
				/* check which element is on the line, nelem is atomic number on C scale */
				nelem = GetElem(chCard);
				iso.n_HighestResolved_max[ipISO][nelem] = numLevels;
				iso_update_num_levels( ipISO, nelem );
			}
			else
			{
				/* keyword element did not appear, so set all species to this number */
				for( nelem=ipISO; nelem<LIMELM; ++nelem )
				{
					iso.n_HighestResolved_max[ipISO][nelem] = numLevels;
					iso_update_num_levels( ipISO, nelem );
				}
			}
		}
	}

	else if( nMatch("ERRO",chCard) && nMatch("GENE" , chCard ) )
	{
		/* Rates will be modified by a randomly generated error that falls within
		 * the range specifically set for each rate (or set of rates).	*/
		iso.lgRandErrGen[ipISO] = true;
		i = 5;
		iso.modelRank[ipISO] = (int)FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL);

		iso.modelRank[ipISO] = MAX2( 0, iso.modelRank[ipISO] );
		if( lgEOL )
			/* Changed to avoid lint complaint */
			/* iso.modelRank[ipISO] = (unsigned)time(NULL); */
			iso.modelRank[ipISO] = abs((int)time(NULL));

		/* this allows a seed that is dependent upon the processor rank
		 * in a parallel run. */
		/* We add 2 so that the seed is always greater than 1, which would reset the generator. */
		init_genrand( (unsigned)iso.modelRank[ipISO] + 2);
	}

	else if( nMatch(" FSM",chCard) )
	{
		if( ipISO == ipH_LIKE )
		{
			fprintf(ioQQQ," Sorry, but fine-structure mixing can only be implemented for the He-like sequence.\n");
			cdEXIT(EXIT_FAILURE);
		}

		/* turn on fine structure mixing of spontaneous decays.  
		 * >>refer	Helike	FSM	Bauman, R., MacAdams, K., and Ferland, G. (2003).	*/
		if( nMatch(" OFF",chCard) )
			iso.lgFSM[ipISO] = false;
		else
			iso.lgFSM[ipISO] = true;
	}

	else if( nMatch("GBAR",chCard) )
	{
		if( ipISO == ipH_LIKE )
		{
			fprintf(ioQQQ," Sorry, the GBAR option is only implemented for the He-like sequence.\n");
			cdEXIT(EXIT_FAILURE);
		}

		/* the HEGBAR command - to change cs of higher levels */
		/* first turn all off, one will be turned back on */
		iso.lgCS_Vriens[ipISO] = false;
		iso.lgCS_None[ipISO] = false;
		iso.nCS_new[ipISO] = false;

		/* now turn one on */
		if( nMatch("VRIE",chCard) )
		{
			/* option to change how collisions for unknown levels affect things */
			iso.lgCS_Vriens[ipISO] = true;
		}
		else if( nMatch(" NEW",chCard) )
		{
			/* option to change how collisions for unknown levels affect things */
			i = 5;
			iso.nCS_new[ipISO] = (int)FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL);

			/* there are two options, 1 and 2, for which fit - default (no number)
			 * will be 1, the broken power law fit */
			if( lgEOL )
				iso.nCS_new[ipISO] = 1;

			ASSERT( iso.nCS_new[ipISO] );
		}
		else if( nMatch(" OFF",chCard) )
		{
			/* option to change how collisions for unknown levels affect things */
			iso.lgCS_None[ipISO] = true;
		}
		else
		{
			fprintf( ioQQQ, " needs parameter\n" );
			cdEXIT(EXIT_FAILURE);
		}
	}


	else if( nMatch("LYMA",chCard) )
	{
		if( ipISO == ipH_LIKE && nMatch("PUMP",chCard) )
		{
			/* >>chng 05 jul 08, separate out Lyman pump commands */
			if( nMatch(" OFF",chCard) )
			{
				/* option to turn off all continuum pumping of Lyman lines */
				hydro.lgLymanPumping = false;
			}
			else if( nMatch("SCALE",chCard) )
			{
				/* multiplicative factor for all continuum pumping of H I Lyman lines,
				 * account for possible emission in the line - only affects H I
				 * not entire H-like iso sequence */
				i = 5;
				hydro.xLymanPumpingScaleFactor = 
					(realnum)FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL);
				/* scale factor is log if negative */
				if( hydro.xLymanPumpingScaleFactor < 0. )
					hydro.xLymanPumpingScaleFactor = 
						(realnum)pow((realnum)10.f , hydro.xLymanPumpingScaleFactor );
			}
			else
			{
				fprintf(ioQQQ," Sorry, I didn\'t recognize an option on this ATOM H-LIKE LYMAN PUMP command.\n");
				fprintf(ioQQQ," The options are \" OFF\", and \"SCALE\".\n");  
				cdEXIT(EXIT_FAILURE);
			}
		}
		else
		{
			/* option to set number of "extra" Lyman lines, used for optical depths only */
			i = 5;
			iso.nLyman[ipISO] = (long int)FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL);
			if( lgEOL )
				NoNumb(chCard);
		}
	}

	/* don't interpolate on look-up tables but compute recombination on the fly instead */
	else if( nMatch("RECO",chCard) &&
		nMatch(" NO ",chCard) && nMatch("INTE",chCard) )
	{
		/* flag set by atom xx-like no recombination interp command, 
		 * says to generate recombination coefficients
		 * on the fly */
		iso.lgNoRecombInterp[ipISO] = true;
	}

	else if( nMatch("REDI",chCard) )
	{
		int ipRedis=0;
		/* there are three functions, PRD_, CRD_, and CRDW,
		 * representing partial redistribution, 
		 * complete redistribution (doppler core only, no wings)
		 * and complete with wings */
		/* partial redistribution */
		if( nMatch(" PRD",chCard) )
		{
			ipRedis = ipPRD;
		}
		/* complete redistribution no wings */
		else if( nMatch(" CRD",chCard) )
		{
			ipRedis = ipCRD;
		}
		/* complete redistribution with wings */
		else if( nMatch("CRDW",chCard) )
		{
			ipRedis = ipCRDW;
		}

		/* if not SHOW option (handled below) then we have a problem */
		else if( !nMatch("SHOW",chCard) )
		{
			fprintf(ioQQQ," There should have been a second keyword on this command.\n");
			fprintf(ioQQQ," Options are _PRD, _CRD, CRDW (_ is space).  Sorry.\n");
			cdEXIT(EXIT_FAILURE);
		}

		/* resonance lines - not Lya*/
		if( nMatch("ALPH",chCard) )
		{
			iso.ipLyaRedist[ipISO] = ipRedis;
		}
		/* Lya itself */
		if( nMatch("RESO",chCard) )
		{
			iso.ipResoRedist[ipISO] = ipRedis;
		}
		/* subordinate lines */
		else if( nMatch("SUBO",chCard) )
		{
			iso.ipSubRedist[ipISO] = ipRedis;
		}
		/* the show option, say what we are assuming */
		else if( nMatch("SHOW",chCard) )
		{
			fprintf(ioQQQ," Ly a is ");
			if( iso.ipLyaRedist[ipISO] ==ipCRDW )
			{
				fprintf(ioQQQ,"complete redistribution with wings\n");
			}
			else if( iso.ipLyaRedist[ipISO] ==ipCRD )
			{
				fprintf(ioQQQ,"complete redistribution with core only.\n");
			}
			else if( iso.ipLyaRedist[ipISO] ==ipPRD )
			{
				fprintf(ioQQQ,"partial redistribution.\n");
			}
			else if( iso.ipLyaRedist[ipISO] ==ipLY_A )
			{
				fprintf(ioQQQ,"special Lya.\n");
			}
			else
			{
				fprintf(ioQQQ," PROBLEM Impossible value for iso.ipLyaRedist.\n");
				TotalInsanity();
			}

			fprintf(ioQQQ," Other %s resonance lines are ",
				elementnames.chElementSym[ipISO] );

			if( iso.ipResoRedist[ipISO] ==ipCRDW )
			{
				fprintf(ioQQQ,"complete redistribution with wings\n");
			}
			else if( iso.ipResoRedist[ipISO] ==ipCRD )
			{
				fprintf(ioQQQ,"complete redistribution with core only.\n");
			}
			else if( iso.ipResoRedist[ipISO] ==ipPRD )
			{
				fprintf(ioQQQ,"partial redistribution.\n");
			}
			else
			{
				fprintf(ioQQQ," PROBLEM Impossible value for iso.ipResoRedist.\n");
				TotalInsanity();
			}

			fprintf(ioQQQ," %s subordinate lines are ",
				elementnames.chElementSym[ipISO] );

			if( iso.ipSubRedist[ipISO] ==ipCRDW )
			{
				fprintf(ioQQQ,"complete redistribution with wings\n");
			}
			else if( iso.ipSubRedist[ipISO] ==ipCRD )
			{
				fprintf(ioQQQ,"complete redistribution with core only.\n");
			}
			else if( iso.ipSubRedist[ipISO] ==ipPRD )
			{
				fprintf(ioQQQ,"partial redistribution.\n");
			}
			else
			{
				fprintf(ioQQQ," PROBLEM Impossible value for iso.ipSubRedist.\n");
				TotalInsanity();
			}
		}
		else
		{
			fprintf(ioQQQ," here should have been another keyword on this command.\n");
			fprintf(ioQQQ," Options are ALPHA, RESONANCE, SUBORDINATE.  Sorry.\n");
			cdEXIT(EXIT_FAILURE);
		}
	}

	/* which type of matrix solution, populations or lowt */
	else if( nMatch("MATR",chCard) )
	{
		if( nMatch(" POP",chCard) )
		{
			strcpy( iso.chTypeAtomSet[ipISO] , "POPU" );
		}
		else if( nMatch(" LOW",chCard) )
		{
			strcpy( iso.chTypeAtomSet[ipISO] , "LOWT" );
		}
		else
		{
			fprintf( ioQQQ, " There should have been a keyword on the MATRIX option.\n");
			fprintf( ioQQQ, " The keywords are POPulations, and LOWte.\n");
			cdEXIT(EXIT_FAILURE);
		}
	}

	else if( nMatch("TOPO",chCard) )
	{
		if( ipISO == ipH_LIKE )
		{
			fprintf( ioQQQ, " The behavior of the TOPOFF option for the H-like sequence has changed since version 07.02.\n");
			fprintf( ioQQQ, " Originally, it accepted a number and distributed \"topoff\" equally into that many levels at the top of the model.\n");
			fprintf( ioQQQ, " Now, the option works the same as in the He-like sequence, which is to turn off topoff altogether.\n" );
		}

		if( nMatch(" OFF",chCard) )
		{
			iso.lgTopoff[ipISO] = false;
			fprintf( ioQQQ, "ISO %li TOPOFF is OFF\n", ipISO );
		}
		else
			iso.lgTopoff[ipISO] = true;
	}


	else
	{
		fprintf( ioQQQ, " There should have been a keyword - STOP\n" );
		cdEXIT(EXIT_FAILURE);
	}
	return;
}
