/* This file is part of Cloudy and is copyright (C)1978-2008 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
/*ParseDont parse the dont command - do not do something */
#include "cddefines.h"
#include "taulines.h"
#include "opacity.h"
#include "phycon.h"
#include "secondaries.h"
#include "pressure.h"
#include "prt.h"
#include "coolheavy.h"
#include "thermal.h"
#include "rfield.h"
#include "dynamics.h"
#include "mole.h"
#include "rt.h"
#include "hmi.h"
#include "atomfeii.h"
#include "yield.h"
#include "ionbal.h"
#include "atmdat.h"
#include "grainvar.h"
#include "hyperfine.h"
#include "parse.h"

void ParseDont(char *chCard )
{
	DEBUG_ENTRY( "ParseDont()" );

	if( (nMatch( "21CM" , chCard ) || nMatch( "21 CM" , chCard )) && 
		nMatch( " LYA" , chCard ) && nMatch( "PUMP" , chCard ) )
	{
		/* no Lya 21 cm pump turns off 21 cm pumping of Lya */
		hyperfine.lgLya_pump_21cm = false;
	}
	else if( nMatch("ADVE",chCard) )
	{
		/*  turn off different aspects of advection */
		if( nMatch("H-LI",chCard) )
		{
			/* advection for the H-like ion sequence */
			dynamics.lgISO[ipH_LIKE] = 0;
		}
		else if( nMatch("HE-L",chCard) )
		{
			/* advection for the He-like ion sequence */
			dynamics.lgISO[ipHE_LIKE] = 0;
		}
		else if( nMatch("META",chCard) )
		{
			/* advection for the everything else - those done in ion_solver */
			dynamics.lgMETALS = 0;
		}
		else if( nMatch("COOL",chCard) )
		{
			/* turn off cooling - heating due to  advection */
			dynamics.lgCoolHeat = 0;
		}
		else
		{
			/* no sub option, so turn them all off */
			dynamics.lgISO[ipH_LIKE] = 0;
			dynamics.lgISO[ipHE_LIKE] = 0;
			dynamics.lgMETALS = 0;
		}

	}

	else if( nMatch("AUGE",chCard) )
	{
		/*  turn off auger effect by killing its block data */
		t_yield::Inst().kill_yield();
		phycon.lgPhysOK = false;
	}

	else if( nMatch("BUFF",chCard) )
	{
		/* NO BUFFERING turn off buffered io for standard output, 
		 * used to get output when code crashes */

		/* >>chng 06 jun 28, moved handling of NO BUFFERING command to cdRead, PvH */

		/* stderr may be a preprocessor macro, so lets be really careful here */
		FILE *test = stderr;
		if( ioQQQ != test )
		{
			/* this should really say stdout and not stderr ! */
			fprintf( ioQQQ, " ioQQQ did not point to stdout, ignored NO BUFFERING command.\n" );
		}

	}

	else if( nMatch("CHAR",chCard) )
	{
		/* turn off all charge transfer interactions */
		atmdat.lgCTOn = false;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("CTHE",chCard) )
	{
		/* turn off charge transfer heating */
		atmdat.HCharHeatOn = 0.;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("COMP",chCard) )
	{
		/* turn off both recoil ionization and compton heating of free electron */
		rfield.comoff = 0.;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("FEII",chCard) )
	{
		/* turn off feii ly-alpha pumping - rate evaluated in FeIILyaPump */
		FeII.lgLyaPumpOn = false;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("FILE",chCard) && nMatch("OPAC",chCard) )
	{
		/* no file opacities, generate them on the fly even if file present */
		opac.lgUseFileOpac = false;
	}

	else if( nMatch("FINE",chCard) && nMatch("OPAC",chCard) )
	{
		/* no fine opacities */
		rfield.lgOpacityFine = false;
	}

	else if( nMatch("FINE",chCard) )
	{
		/* turn off fine structure optical depths */
		rt.lgFstOn = false;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("FREE",chCard) )
	{
		/* turn off free free heating and cooling */
		CoolHeavy.lgFreeOn = false;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("GRAI",chCard) )
	{
		if( nMatch("NEUT",chCard) )
		{
			/* turn off ion grain recombination "NO GRAIN NEUTRALIZATION" */
			ionbal.lgGrainIonRecom = false;
			phycon.lgPhysOK = false;
		}
		else if( nMatch("GAS ",chCard) && nMatch("COLL",chCard) && nMatch("ENER",chCard) )
		{
			/* turn off grain - gas collisional energy exchange 
			 * "NO GRAIN GAS COLLISIONAL ENERGY EXCHANGE " */
			gv.lgDColOn = false;
			phycon.lgPhysOK = false;
		}
		else if( nMatch("ELEC",chCard) )
		{
			/* turn off grain contributions to electrons "NO GRAIN ELECTRONS" */
			gv.lgGrainElectrons = false;
			phycon.lgPhysOK = false;
		}
		else if( nMatch("MOLE",chCard) )
		{
			/* turn off capture of molecules on grain surfaces "NO GRAIN MOLECULES" */
			mole.lgGrain_mole_deplete = false;
			phycon.lgPhysOK = false;
		}
		else if( nMatch("QHEA",chCard) )
		{
			/* turn off quantum heating of grains "NO GRAIN QHEAT" */
			gv.lgQHeatOn = false;
			phycon.lgPhysOK = false;
		}
		else if( nMatch("X-RA",chCard) )
		{
			/* revert to WD01 physics "NO GRAIN X-RAY TREATMENT" */
			gv.lgWD01 = true;
		}
		else if( nMatch("PHYSICS",chCard) )
		{
			/* turn off grain physics "NO GRAIN PHYSICS" */
			gv.lgGrainPhysicsOn = false;
			phycon.lgPhysOK = false;
		}
		else
		{
			fprintf( ioQQQ, " No key recognized on this line.\n" );
			cdEXIT(EXIT_FAILURE);
		}
	}

	/* no induced processes */
	else if( nMatch("INDU",chCard) )
	{
		/* turn off induced recombination, stimulated emission,
		 * continuum fluorescent excitation of lines, 
		 * stimulated emission correction to optical depths attenuation */
		rfield.lgInducProcess = false;
	}

	/* no collisional ionization */
	else if( nMatch("COLL",chCard) && nMatch("IONI",chCard) )
	{
		/* turn off collisional ionization */
		atmdat.lgCollIonOn = false;
	}

	else if( nMatch("LEVE",chCard) )
	{
		/* turn off the set of level 2 lines, safe for lower densities 
		 * this is the upper limit to the counter that is always used,
		 * so no loops will ever occur */
		/* >>chng 06 mar 04 from -1 to 0 so that size_t is zero in state put & get */
		/*nWindLine = -1;*/
		nWindLine = 0;
	}

	/* various no line options */
	else if( nMatch("LINE",chCard) && !nMatch(" OTS",chCard) )
	{
		if( nMatch("DIFF",chCard) && nMatch("PUMP",chCard) )
		{
			/* no diffuse line pumping, 
			* turn off pumping of lines by diffuse continuum*/
			rfield.DiffPumpOn = 0.;
		}
		else if( nMatch("TRAN",chCard) )
		{
			/* no line transfer command */
			rfield.lgDoLineTrans = false;
		}
		else
		{
			/* missing no line option */
			fprintf( ioQQQ, " There has to be an option on the NO LINE command.\n" );
			fprintf( ioQQQ, " The optinoas are DIFFUSE PUMP and TRANSFER.\n Sorry.\n" );
			cdEXIT(EXIT_FAILURE);
		}
	}


	else if( nMatch("OPAC",chCard) && nMatch("REEVAL",chCard) )
	{
		/* don't constantly reevaluate the opacities */
		rfield.lgOpacityReevaluate = false;
	}

	else if( nMatch("IONI",chCard) && nMatch("REEVAL",chCard) )
	{
		/* "no ionization reevaluation" - don't constantly reevaluate the ionization */
		rfield.lgIonizReevaluate = false;
	}

	/* options to kill ots components as debugging aids */
	else if( nMatch(" OTS",chCard) )
	{
		if( nMatch(" LYA",chCard) )
		{
			/* turn off Lya ots rates - for debugging oscillations */
			rfield.lgLyaOTS = false;
		}

		else if( nMatch("HEII",chCard) )
		{
			/* turn off Lya ots rates - for debugging oscillations */
			rfield.lgHeIIOTS = false;
		}

		else if( nMatch("LINE",chCard) )
		{
			/* turn off line ots rates - for debugging oscillations */
			rfield.lgKillOTSLine = true;
		}
	}

	/* options to kill outward compoents as a debugging aid */
	else if( nMatch("OUTW",chCard) )
	{
		if( nMatch("LINE",chCard) )
		{
			/* turn off Lya ots rates - for debugging oscillations */
			rfield.lgKillOutLine = true;
		}
		else if( nMatch("CONT",chCard) )
		{
			/* turn off Lya ots rates - for debugging oscillations */
			rfield.lgKillOutCont = true;
		}
	}
	else if( nMatch("MOLE",chCard) )
	{
		/* disable molecule formation, first option is to turn off only H2 part */
		if( nMatch(" H2 ",chCard) )
		{
			/* turn off only H2 */
			hmi.lgNoH2Mole = true;
		}
		else if( nMatch(" CO ",chCard) )
		{
			/* turn off CO network */
			co.lgNoCOMole = true;
		}
		else
		{
			/* turn off both CO and H2 networks */
			hmi.lgNoH2Mole = true;
			co.lgNoCOMole = true;
		}
		phycon.lgPhysOK = false;
	}

	else if( nMatch("PHOT",chCard) )
	{
		/* disable photoionization */
		ionbal.lgPhotoIoniz_On = false;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("RADI",chCard) )
	{
		/* don't include line radiation pressure */
		pressure.lgLineRadPresOn = false;
	}

	else if( nMatch("RECO",chCard) )
	{
		/* disable compton recoil of bound electrons - "no recoil ioniz" */
		ionbal.lgCompRecoil = false;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("SCAT",chCard) && nMatch("OPAC",chCard))
	{
		/* no scattering opacity, for Compton thick spherical geometry */
		opac.lgScatON = false;
	}

	else if( nMatch("SCAT",chCard) && nMatch("ESCA",chCard))
	{
		/* no electron scattering contribution to line escape probs */
		rt.lgElecScatEscape = false;
	}

	else if( nMatch("SECO",chCard) )
	{
		/* turn off secondary electron ionizations */
		secondaries.lgSecOFF = true;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("SPOT",chCard) )
	{
		/* no on-the-spot; turn on all ground state rec */
		opac.otsmin = 1.;
	}

	else if( nMatch("STAR",chCard) )
	{
		/* no stark broadening */
		rt.lgStarkON = false;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("STAT",chCard) )
	{
		/* no static opacities - constantly reevaluate them */
		opac.lgOpacStatic = false;
	}

	else if( nMatch("TEPR",chCard) )
	{
		/* no tepredictor */
		/* turn off prediction of next zone's temperature, as guessed in ZoneStart */
		thermal.lgPredNextTe = false;
	}

	else if( nMatch("THRE",chCard) )
	{
		/* turn off Cota's three body rec subroutine */
		ionbal.lgNoCota = true;
		phycon.lgPhysOK = false;
	}

	else if( nMatch("TIME",chCard) )
	{
		/* don't print anything with a time, so that we can expect
		 * perfect agreement between separate runs */
		prt.lgPrintTime = false;
	}

	else if( nMatch(" UTA",chCard) )
	{
		/* turn off ALL inner shell absorption ionization */
		ionbal.lgInnerShellLine_on = false;
		phycon.lgPhysOK = false;
	}

	/* the no vary command is parsed well before we get to this point,
	 * but we have to do something here or the parser will say that
	 * no command existed on the command line */
	else if( strncmp(chCard , "NO VARY" , 7 ) == 0 )
	{
		/* this is a no-nothing, picked up to stop optimizer */
		((void)0);
	}

	else
	{
		/* end of else if trap */
		fprintf( ioQQQ, 
			" I do not recognize a keyword on this NO ... command.\n %s\n Sorry.\n" , 
			chCard);
		cdEXIT(EXIT_FAILURE);
	}

	/* this option, if keyword (OK) appears, then do not set warning */
	if( nMatch("(OK)",chCard) )
	{
		/* say that physical conditions are actually ok */
		phycon.lgPhysOK = true;
	}
	return;
}
