/* This file is part of Cloudy and is copyright (C)1978-2008 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
/*ParseHDEN parse the hden command */
#include "cddefines.h"
#include "input.h"
#include "dense.h"
#include "optimize.h"
#include "parse.h"

void ParseHDEN(char *chCard )
{
	bool lgEOL;
	long int i;

	DEBUG_ENTRY( "ParseHDEN()" );

	if( dense.gas_phase[ipHYDROGEN] > 0. )
	{
		fprintf( ioQQQ, " PROBLEM DISASTER More than one density command was entered.\n" );
		cdEXIT(EXIT_FAILURE);
	}

	/* log of hydrogen density */
	i = 5;
	dense.gas_phase[ipHYDROGEN] = (realnum)FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL);
	if( lgEOL )
	{
		fprintf( ioQQQ, " DISASTER The density MUST be entered with this command. STOP\n" );
		cdEXIT(EXIT_FAILURE);
	}

	/* check for further options */
	if( ! nMatch("LINE",chCard) )
	{
		/* check size of density - will we crash? */
		if( dense.gas_phase[ipHYDROGEN] > log10(FLT_MAX) ||
		    dense.gas_phase[ipHYDROGEN] < log10(FLT_MIN) )	
		{
			fprintf(ioQQQ,
				" DISASTER - the log of the entered hydrogen density is %.3f - too extreme for this processor.\n",
				dense.gas_phase[ipHYDROGEN]);
			if( dense.gas_phase[ipHYDROGEN] > 0. )
				fprintf(ioQQQ,
					" DISASTER - the log of the largest hydrogen density this processor can do is %.3f.\n",
					log10(FLT_MAX) );
			else
				fprintf(ioQQQ,
					" DISASTER - the log of the smallest hydrogen density this processor can do is %.3f.\n",
					log10(FLT_MIN) );
			fprintf(ioQQQ," Sorry.\n" );
			
			cdEXIT(EXIT_FAILURE);
		}

		dense.gas_phase[ipHYDROGEN] = (realnum)pow((realnum)10.f,dense.gas_phase[ipHYDROGEN]);
	}

	if( dense.gas_phase[ipHYDROGEN] <= 0. )
	{
		fprintf( ioQQQ, " PROBLEM DISASTER Hydrogen density must be > 0.\n" );
		cdEXIT(EXIT_FAILURE);
	}

	/* this is the linear initial density */
	dense.den0 = dense.gas_phase[ipHYDROGEN];

	/* check if exponent entered */
	dense.DensityPower = (realnum)FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL);
	/* this branch when exponent was entered - do something with it */
	if( !lgEOL )
	{
		/*  not constant density
		 *  some sort of power law density distribution */
		if( nMatch("COLU",chCard) )
		{
			/* density will depend on column density to a power
			 * number entered is col den, convert to scale radius
			 * at this point HDEN is LOG10 of hydrogen density */
			dense.rscale = (realnum)pow(10.,FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL));
			if( lgEOL )
			{
				fprintf( ioQQQ, " The column density MUST be set if the col den option is to be used.\n" );
				cdEXIT(EXIT_FAILURE);
			}
			strcpy( dense.chDenseLaw, "POWC" );
		}
		else if( nMatch("DEPT",chCard) )
		{
			/* depth option, sets scale radius, log cm */
			dense.rscale = (realnum)pow(10.,FFmtRead(chCard,&i,INPUT_LINE_LENGTH,&lgEOL));
			if( lgEOL )
			{
				fprintf( ioQQQ, " The scale depth MUST be set if the depth option is to be used.\n" );
				cdEXIT(EXIT_FAILURE);
			}
			strcpy( dense.chDenseLaw, "POWD" );
		}
		else
		{
			/* radius option, will be relative to inner radius */
			strcpy( dense.chDenseLaw, "POWR" );
		}
	}

	/* vary option */
	if( optimize.lgVarOn )
	{
		/*  pointer to where to write */
		optimize.nvfpnt[optimize.nparm] = input.nRead;
		optimize.vparm[0][optimize.nparm] = (realnum)log10(dense.gas_phase[ipHYDROGEN]);
		optimize.vincr[optimize.nparm] = 1.;

		/* these are varios options for density laws, 
		 * first is constant density or pressre*/
		if( strcmp(dense.chDenseLaw ,"CDEN") == 0 || 
			 strcmp(dense.chDenseLaw ,"CPRE") == 0 )
		{
			strcpy( optimize.chVarFmt[optimize.nparm], "HDEN=%f" );
			optimize.nvarxt[optimize.nparm] = 1;
		}

		/* power law density distrution */
		else if( strcmp(dense.chDenseLaw,"POWR") == 0 )
		{
			strcpy( optimize.chVarFmt[optimize.nparm], "HDEN=%f, power=%f" );
			optimize.vparm[1][optimize.nparm] = dense.DensityPower;
			optimize.nvarxt[optimize.nparm] = 2;
		}

		/* power law with density scale depending on column density */
		else if( strcmp(dense.chDenseLaw,"POWC") == 0 )
		{
			strcpy( optimize.chVarFmt[optimize.nparm], "HDEN=%f, power=%f, column=%f" );
			optimize.nvarxt[optimize.nparm] = 3;
			optimize.vparm[1][optimize.nparm] = dense.DensityPower;
			optimize.vparm[2][optimize.nparm] = (realnum)log10(dense.rscale);
		}

		/* power law with density scale depending on depth */
		else if( strcmp(dense.chDenseLaw,"POWD") == 0 )
		{
			strcpy( optimize.chVarFmt[optimize.nparm], "HDEN=%f, power=%f, depth=%f" );
			optimize.nvarxt[optimize.nparm] = 3;
			optimize.vparm[1][optimize.nparm] = dense.DensityPower;
			optimize.vparm[2][optimize.nparm] = (realnum)log10(dense.rscale);
		}

		/* could not idensity an option */
		else
		{
			fprintf( ioQQQ, " Internal error in HDEN\n" );
			cdEXIT(EXIT_FAILURE);
		}
		++optimize.nparm;
	}
	return;
}
