/* This file is part of Cloudy and is copyright (C)1978-2008 by Gary J. Ferland and
* others.  For conditions of distribution and use see copyright notice in license.txt */
/*ParseTest parse the test command */
#include "cddefines.h"
#include "atomfeii.h"
#include "input.h"
#include "assertresults.h"
#include "parse.h"

void ParseTest(char *chCard , 
			   long int *nqh,
			   realnum *ar1,
			   bool lgDSet)
{
	char chStuff[INPUT_LINE_LENGTH];

	DEBUG_ENTRY( "ParseTest()" );

	/* do a smoke test - 
	* generate test input stream for rapid testing of code */

	/* optional keyword PRINT will cause all the commands to be printed */
	int nPrintTest = nMatch("PRIN" , chCard );

	/* option to turn on the H2 molecule */
	if( nMatch(" H2 ",chCard) )
	{
		/* this will both enable the molecule and also set the threshold for actually
		* computing it to a very low value, so that it is actually used in ionized gas */
		sprintf( chStuff , "ATOM H2 LIMIT -20 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseAtomH2(chStuff);
	}

	/* this is option to also turn on large FeII atom */
	if( nMatch("FEII",chCard) || nMatch("FE II",chCard) )
	{
		sprintf( chStuff , "ATOM FEII " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseAtomFeII(chStuff);
	}

	/* use largest possible hydrogen atom */
	if( nMatch("LARG",chCard) )
	{
		sprintf( chStuff , "ATOM H-LIKE ELEMENT HYDROGEN LEVELS LIMIT  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseAtomISO(ipH_LIKE,chStuff);
	}

	if( nMatch("MOLE",chCard) )
	{
#		if 0
		/* hydrogen density */
		sprintf( chStuff , "TRACE TEMPERATURE CONVERGENCE " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseTrace(chStuff);
#		endif

		/* hydrogen density */
		sprintf( chStuff , "HDEN 5 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseHDEN(chStuff);

		/* make a constant temperature model */
		sprintf( chStuff , "CONSTANT TEMPER 50K  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseConstant(chStuff);

		/* continuum to include full energy range */
		sprintf( chStuff , "TABLE ISM  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseTable( nqh,chStuff, ar1);

		/* extinguish this continuum */
		sprintf( chStuff , "EXTINGUISH 23  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseExtinguish( chStuff );

		/* stop in second zone, so we do use the zone increment logic */
		sprintf( chStuff , "STOP ZONE 2  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseStop(chStuff);

		/* set thickness */
		sprintf( chStuff , "SET DR 0  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseSet(chStuff);

		/* do Case B so lyman line pumping of H is not important */
		sprintf( chStuff , "CASE B  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseCaseB(chStuff);

		/* set cosmic rays */
		sprintf( chStuff , "COSMIC RAY BACKGROUND  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseCosmicRays(chStuff);

		/* add grains and ism abundances */
		sprintf( chStuff , "ABUNDANCES ISM  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseAbundances(chStuff , lgDSet);

		/* add grains and ism abundances */
		sprintf( chStuff , "CONSTANT GRAIN TEMPERATURE 20K  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseConstant(chStuff);

		/* create series of assert commands */
		sprintf( input.chCARDCAPS , "ASSERT EDEN 0.528 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/* create series of assert commands */
		sprintf( input.chCARDCAPS , "ASSERT MOLECULAR FRACTION H2 -1.193 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/* create series of assert commands */
		sprintf( input.chCARDCAPS , "ASSERT COLUMN CO 1.214 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/* create series of assert commands */
		sprintf( input.chCARDCAPS , "ASSERT EDEN 0.528 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/* create series of assert commands */
		sprintf( input.chCARDCAPS , "ASSERT HYDROGEN 1 TEMPERATURE 50K " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();
	}

	else
	{
		/* hydrogen density */
		sprintf( chStuff , "HDEN 4 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseHDEN(chStuff);

		/* make a constant temperature model */
		sprintf( chStuff , "CONSTANT TEMPER 4  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseConstant(chStuff);

		/* continuum to include full energy range */
		sprintf( chStuff , "TABLE AGN  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseTable( nqh,chStuff, ar1);

		/* set ionization parameter */
		sprintf( chStuff , "IONIZATION PARAMETER -2  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseIonPar( nqh,chStuff, ar1);

		/* use old abundances */
		sprintf( chStuff , "ABUNDANCES OLD SOLAR 84  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseAbundances(chStuff,lgDSet);

		/* >>chng 02 apr 19, add this */
		/* set total Lyman continuum depth - this is to prevent caution
		* that Lyman continuum was thin but expected to be thick */
		sprintf( chStuff , "STOP LYMAN OPTICAL -4  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseStop(chStuff);

		/* stop in second zone, so we do use the zone increment logic */
		sprintf( chStuff , "STOP ZONE 2  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseStop(chStuff);

		/* set thickness */
		sprintf( chStuff , "SET DR 0  " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , chStuff );
		ParseSet(chStuff);

		/* create series of assert commands */
		sprintf( input.chCARDCAPS , "ASSERT HYDROGEN 1 IONIZATION -3.040 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		sprintf( input.chCARDCAPS , "ASSERT HELIUM 2 IONIZATION -1.067 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		sprintf( input.chCARDCAPS , "ASSERT CARBON 2 IONIZATION -2.301 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/*>>chng 06 dec 01, from -0.653 to -0.560. Badnell DR by default */
		sprintf( input.chCARDCAPS , "ASSERT CARBON 3 IONIZATION -0.560 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/*>>chng 06 dec 01, from -0.348 to -0.373. Badnell DR by default */
		sprintf( input.chCARDCAPS , "ASSERT CARBON 4 IONIZATION -0.373 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/*>>chng 06 dec 01, from -0.490 to -0.530. Badnell DR by default */
		sprintf( input.chCARDCAPS , "ASSERT CARBON 5 IONIZATION -0.530 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/*>>chng 06 dec 01, from -0.800 to -0.861. Badnell DR by default */
		sprintf( input.chCARDCAPS , "ASSERT OXYGEN 3 IONIZATION  -0.861 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/*>>chng 06 dec 01, from -0.180 to -0.157. Badnell DR by default */
		sprintf( input.chCARDCAPS , "ASSERT OXYGEN 4 IONIZATION -0.157 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/*>>chng 06 dec 01, from -0.770 to -0.808. Badnell DR by default */
		sprintf( input.chCARDCAPS , "ASSERT OXYGEN 5 IONIZATION -0.808 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();

		/* >>chng 02 apr 19, from 0.7258 to 0.946, due to adding Lyman cont depth */
		/* >>chng 07 oct 22, from 0.946  to 1.108, resolve l-levels of h-like sequence */
		sprintf( input.chCARDCAPS , "ASSERT LINE \"CA B\" 4861 1.108 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		/* must have copy of this in chOrgCard, which is used by the routine to get lab */
		strcpy( input.chOrgCard , input.chCARDCAPS);
		ParseAssertResults();

		/* >>chng 02 apr 19, from 2.4603 to 3.25, due to adding Lyman cont depth 
		 * >>chng 06 nov 17, asserted value from 3.25 to 3.11, drift over last year */
		/* >>chng 06 dec 01, from 3.11 to 2.72. Badnell DR by default */
		/* >>chng 07 oct 22, from 2.72 to 3.18, resolve l-levels of h-like sequence */
		sprintf( input.chCARDCAPS , "ASSERT LINE \"O  3\" 5007 3.18 " );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		/* must have copy of this in chOrgCard, which is used by the routine to get lab */
		strcpy( input.chOrgCard , input.chCARDCAPS);
		ParseAssertResults();

		sprintf( input.chCARDCAPS , "ASSERT HTOT -15.011" );
		if( nPrintTest )
			fprintf(ioQQQ , "%s\n" , input.chCARDCAPS );
		ParseAssertResults();
	}
	return;
}
