/* This file is part of Cloudy and is copyright (C)1978-2008 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */

#ifndef _VERSION_H_
#define _VERSION_H_

/** version.h */
#include "date.h"

static const int CLD_MAJOR = 8;
static const int CLD_MINOR = 0;
static const int CLD_PATCH = 0;

static const string Url = "$HeadURL$";

static const char* CITATION = "Ferland, G. J., Korista, K. T., Verner, D. A., Ferguson, J. W., Kingdon, J. B., & Verner, E. M. 1998, PASP, 110, 761";
static const char* CITATION_LATEX = "\\bibitem[Ferland et al.(1998)]{1998PASP..110..761F} Ferland, G.~J., Korista, K.~T., Verner, D.~A., Ferguson, J.~W., Kingdon, J.~B., \\& Verner, E.~M.\\ 1998, \\pasp, 110, 761";
static const char* CITATION_SHORT = "Ferland et al. (1998)";		

class t_version : public Singleton<t_version> 
{
	friend class Singleton<t_version>;
protected:
	t_version()
	{
		static const char chMonth[12][4] =
			{ "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

		/* is this a release version? */
		lgRelease = true;

		/* is this a beta version?  0 for no
		 * if this is non-zero then lgRelease above should be false */
		nBetaVer = 0;

		if( lgRelease )
		{
			if( CLD_PATCH > 0 )
				sprintf( chVersion, "%2.2i.%2.2i (patch level %d)", CLD_MAJOR, CLD_MINOR, CLD_PATCH );
			else
				sprintf( chVersion, "%2.2i.%2.2i", CLD_MAJOR, CLD_MINOR );
		}
		else if( nBetaVer > 0 )
		{
			sprintf( chVersion, "%2.2i.%2.2i beta %ld (prerelease)", CLD_MAJOR, CLD_MINOR, nBetaVer );
		}
		else
		{
			sprintf( chVersion, "%2.2i.%2.2i.%2.2i", YEAR%100, MONTH+1, DAY );
		}

		sprintf( chDate, "%2.2i%3.3s%2.2i", YEAR%100, chMonth[MONTH], DAY );

		// analyze the URL to determine where we live, the version number is derived from that
		// the code below is based on the following naming scheme:
		//
		// /branches/c08_branch -- release branch, all bug fixes are submitted here 
		//
		// /tags/develop/c08.00_rc1 -- release candidates go here 
		//
		// /tags/release/c08.00 -- first official release
		// /tags/release/c08.01 -- first bug-fix rollup, etc... 
		//
		// /tags/patch_versions/c08.00_pl00 -- identical to /tags/release/c08.00
		// /tags/patch_versions/c08.00_pl01 -- first patch update, etc... 

		vector<string> Part;
		Split( Url, "/", Part, SPM_RELAX );
		if( Part.size() >= 3 )
		{
			// the last two parts are "source" and "version.h $", we don't need them...
			// the one before is the relevant identifier (e.g. "c08.01")
			// for conciseness we will refer to it below as the "branch"
			string Branch = Part[Part.size()-3];

			bool lgReleaseTag = ( Url.find("/tags/release/") != string::npos );
			bool lgPatchTag = ( Url.find("/tags/patch_versions/") != string::npos );
			bool lgDevelopTag = ( Url.find("/tags/develop/") != string::npos );
			// this expects a branch name like "c08_branch"
			bool lgReleaseBranch = ( Url.find("/branches/") != string::npos &&
						 Branch.size() == 10 && Branch[0] == 'c' &&
						 Branch.find("_branch") != string::npos );

			bool lgReleaseChk = ( lgReleaseTag || lgPatchTag );

			long nBetaVerChk;

			// determine if this is a beta version
			string::size_type ptr;
			if( lgDevelopTag && ( ptr = Branch.find( "_rc" ) ) != string::npos )
				// this expects a branch name like "c08.00_rc1"
				sscanf( Branch.substr( ptr+3 ).c_str(), "%ld", &nBetaVerChk );
			else
				nBetaVerChk = 0;

			int nMajorLevel=0, nMinorLevel=0, nPatchLevel=0;

			if( lgReleaseBranch || lgReleaseChk || nBetaVerChk > 0 )
			{
				// this expects a branch name starting with "c08"
				sscanf( Branch.substr(1,2).c_str(), "%d", &nMajorLevel );
				if( nMajorLevel != CLD_MAJOR )
					fprintf( ioQQQ, "PROBLEM - CLD_MAJOR mismatch, please check version.h\n" );
			}

			if( lgReleaseChk || nBetaVerChk > 0 )
			{
				// this expects a branch name starting with "c08.01"
				sscanf( Branch.substr(4,2).c_str(), "%d", &nMinorLevel );
				if( nMinorLevel != CLD_MINOR )
					fprintf( ioQQQ, "PROBLEM - CLD_MINOR mismatch, please check version.h\n" );
			}
			
			if( lgPatchTag )
			{
				// this expects a branch name like "c08.01_pl02"
				sscanf( Branch.substr(9,2).c_str(), "%d", &nPatchLevel );
				if( nPatchLevel != CLD_PATCH )
					fprintf( ioQQQ, "PROBLEM - CLD_PATCH mismatch, please check version.h\n" );
				// c08.00_pl00 is identical to release c08.00, so pass it off as the latter...
				if( nPatchLevel == 0 )
					lgReleaseTag = true;
			}

			if( nBetaVerChk > 0 )
			{
				if( nBetaVerChk != nBetaVer )
					fprintf( ioQQQ, "PROBLEM - nBetaVer mismatch, please check version.h\n" );
			}

			if( lgReleaseChk != lgRelease )
				fprintf( ioQQQ, "PROBLEM - lgRelease setting wrong, please check version.h\n" );
		}

		char mode[6];
		if( sizeof(long) == 4 )
			strncpy( mode, "ILP32", 6 );
		else if( sizeof(long) == 8 )
			strncpy( mode, "LP64", 6 );
		else
			strncpy( mode, "?????", 6 );

		/* now generate info on how we were compiled, including compiler version */
		sprintf( chInfo,
			 "cdInit compiled on %s in OS %s using the %s %i compiler in %s mode.",
			 __DATE__, __OS, __COMP, __COMP_VER, mode );

		chCitation = CITATION;
		chCitationLatex = CITATION_LATEX;
		chCitationShort = CITATION_SHORT;
	}
public:
	/** date of this version of the code as a string */
	char chDate[INPUT_LINE_LENGTH];

	/** version string of this version of the code */
	char chVersion[INPUT_LINE_LENGTH];

	/** normally zero, non-zero if this is a beta test version */
	long int nBetaVer;

	/** is this a release version?  if so do not print some internal comments */
	bool lgRelease;

	/** information about when and how the code was compiled, including 
	 * compiler version */
	char chInfo[INPUT_LINE_LENGTH];

	/** string to print with print citation command */
	const char *chCitation;

	/** abbreviated form to print with print citation command */
	const char *chCitationShort;

	/** string to print latex bibitem with print citation command */
	const char *chCitationLatex;
};

#endif /* _VERSION_H_ */
